"""Bijections that wrap jax function transforms (scan and vmap)."""
from functools import partial
from typing import Any, Callable

import equinox as eqx
import jax.numpy as jnp
from jax.lax import scan

from flowjax.bijections.bijection import Bijection


class Batch(Bijection):
    """Add batch dimensions to a bijection, such that the new shape is
    batch_shape + bijection.shape. The batch dimensions are added using multiple
    applications of eqx.filter_vmap.
    """

    bijection: Bijection
    in_axes: tuple
    batch_shape: tuple[int, ...]

    def __init__(
        self,
        bijection: Bijection,
        batch_shape: tuple[int, ...],
        vectorize_bijection: bool,
        vectorize_condition: bool | None = None,
    ):
        """
        Args:
            bijection (Bijection): Bijection to add batch dimensions to.
            batch_shape (tuple[int, ...]): The shape of the batch dimension.
            vectorize_bijection (bool): Whether to vectorise bijection parameters.
                If True, we vectorize across the leading dimensions in the array
                leaves of the bijection. In this case, the array leaves must
                have leading dimensions equal to batch_shape. For construction of
                compatible bijections, see ``eqx.filter_vmap``. If False: we broadcast
                the parameters, i.e. the same bijection parameters are used for each x.
            vectorize_condition (bool | None): Whether to vectorize or broadcast the
                conditioning variables. If broadcasting, the condition shape is
                unchanged. If vectorising, the condition shape will be
                ``batch_shape + bijection.cond_shape``. Defaults to None.

        Example:

            .. doctest::

                >>> import jax.numpy as jnp
                >>> from flowjax.bijections import Batch, Affine
                >>> x = jnp.ones(2)
                >>> batched = Batch(Affine(1), (2,), vectorize_bijection=False)
                >>> batched.transform(x)
                Array([2., 2.], dtype=float32)
        """
        if vectorize_condition is None and bijection.cond_shape is not None:
            raise ValueError(
                "vectorize_condition must be specified for conditional bijections."
            )

        self.in_axes = (
            eqx.if_array(axis=0) if vectorize_bijection else None,
            eqx.if_array(axis=0),
            0 if vectorize_condition else None,
        )
        self.shape = batch_shape + bijection.shape
        self.batch_shape = batch_shape
        self.bijection = bijection

        if self.bijection.cond_shape is None:
            self.cond_shape = None
        elif vectorize_condition:
            self.cond_shape = batch_shape + self.bijection.cond_shape
        else:
            self.cond_shape = self.bijection.cond_shape

    def transform(self, x, condition=None):
        x, condition = self._argcheck_and_cast(x, condition)

        def _transform(bijection, x, condition):
            return bijection.transform(x, condition)

        return self.multi_vmap(_transform)(self.bijection, x, condition)

    def transform_and_log_det(self, x, condition=None):
        x, condition = self._argcheck_and_cast(x, condition)

        def _transform_and_log_det(bijection, x, condition):
            return bijection.transform_and_log_det(x, condition)

        y, log_det = self.multi_vmap(_transform_and_log_det)(
            self.bijection, x, condition
        )
        return y, jnp.sum(log_det)

    def inverse(self, y, condition=None):
        y, condition = self._argcheck_and_cast(y, condition)

        def _inverse(bijection, x, condition):
            return bijection.inverse(x, condition)

        return self.multi_vmap(_inverse)(self.bijection, y, condition)

    def inverse_and_log_det(self, y, condition=None):
        y, condition = self._argcheck_and_cast(y, condition)

        def _inverse_and_log_det(bijection, x, condition):
            return bijection.inverse_and_log_det(x, condition)

        x, log_det = self.multi_vmap(_inverse_and_log_det)(self.bijection, y, condition)
        return x, jnp.sum(log_det)

    def multi_vmap(self, func: Callable) -> Callable:
        """Compose vmap to add ndim batch dimensions."""
        for _ in range(len(self.batch_shape)):
            func = eqx.filter_vmap(func, in_axes=self.in_axes)
        return func


class Scan(Bijection):
    """Repeatedly apply the same bijection with different parameter values. Internally,
    uses `jax.lax.scan` to reduce compilation time.
    """

    static: Any
    params: Any

    def __init__(self, bijection: Bijection):
        """
        The array leaves in `bijection` should have an additional leading axis to scan
        over. Often it is convenient to construct these using `equinox.filter_vmap`.

        Args:
            bijection (Bijection): A bijection, in which the arrays leaves have an
                additional leading axis to scan over. For complex bijections, it can be
                convenient to create compatible bijections with ``equinox.filter_vmap``.

        Example:
            Below is equivilent to ``Chain([Affine(p) for p in params])``.

            .. doctest::

                >>> from flowjax.bijections import Scan, Affine
                >>> import jax.numpy as jnp
                >>> import equinox as eqx
                >>> params = jnp.ones((3, 2))
                >>> affine = Scan(Affine(params))

        """
        self.params, self.static = eqx.partition(bijection, eqx.is_array)
        self.shape = bijection.shape
        self.cond_shape = bijection.cond_shape

    def transform(self, x, condition=None):
        x, condition = self._argcheck_and_cast(x, condition)

        def step(x, params, condition=None):
            bijection = eqx.combine(self.static, params)
            result = bijection.transform(x, condition)
            return (result, None)

        step = partial(step, condition=condition)
        y, _ = scan(step, x, self.params)
        return y

    def transform_and_log_det(self, x, condition=None):
        x, condition = self._argcheck_and_cast(x, condition)

        def step(carry, params, condition):
            x, log_det = carry
            bijection = eqx.combine(self.static, params)
            y, log_det_i = bijection.transform_and_log_det(x, condition)  # type: ignore
            return ((y, log_det + log_det_i.sum()), None)

        step = partial(step, condition=condition)
        (y, log_det), _ = scan(step, (x, 0), self.params)
        return y, log_det

    def inverse(self, y, condition=None):
        y, _ = self._argcheck_and_cast(y, condition)

        def step(y, params, condition):
            bijection = eqx.combine(self.static, params)
            x = bijection.inverse(y, condition)  # type: ignore
            return (x, None)

        step = partial(step, condition=condition)
        x, _ = scan(step, y, self.params, reverse=True)
        return x

    def inverse_and_log_det(self, y, condition=None):
        y, _ = self._argcheck_and_cast(y, condition)

        def step(carry, params, condition):
            y, log_det = carry
            bijection = eqx.combine(self.static, params)
            x, log_det_i = bijection.inverse_and_log_det(y, condition)  # type: ignore
            return ((x, log_det + log_det_i.sum()), None)

        step = partial(step, condition=condition)
        (y, log_det), _ = scan(step, (y, 0), self.params, reverse=True)
        return y, log_det



"""Premade versions of common flow architetctures from ``flowjax.flows``."""
# Note that here although we could chain arbitrary bijections using `Chain`, here,
# we generally opt to use `Scan`, which avoids excessive compilation
# when the flow layers share the same structure.

from typing import Callable

import equinox as eqx
import jax.nn as jnn
import jax.numpy as jnp
from equinox.nn import Linear
from jax import random
from jax.nn.initializers import glorot_uniform
from jax.random import KeyArray

from flowjax.bijections import (
    AdditiveCondition,
    Bijection,
    BlockAutoregressiveNetwork,
    Chain,
    Coupling,
    Flip,
    Invert,
    MaskedAutoregressive,
    Permute,
    RationalQuadraticSpline,
    Scan,
    TanhLinearTails,
    TriangularAffine,
)
from flowjax.distributions import Distribution, Transformed


class CouplingFlow(Transformed):
    """Coupling flow (https://arxiv.org/abs/1605.08803)."""

    flow_layers: int
    nn_width: int
    nn_depth: int
    permute_strategy: str

    def __init__(
        self,
        key: KeyArray,
        base_dist: Distribution,
        transformer: Bijection,
        cond_dim: int | None = None,
        flow_layers: int = 8,
        nn_width: int = 40,
        nn_depth: int = 2,
        nn_activation: Callable = jnn.relu,
        invert: bool = True,
    ):
        """
        Args:
            key (KeyArray): Jax PRNGKey.
            base_dist (Distribution): Base distribution.
            transformer (Bijection): Bijection to be parameterised by conditioner.
            cond_dim (int): Dimension of conditioning variables. Defaults to None.
            flow_layers (int): Number of coupling layers. Defaults to 5.
            nn_width (int): Conditioner hidden layer size. Defaults to 40.
            nn_depth (int): Conditioner depth. Defaults to 2.
            nn_activation (int): Conditioner activation function. Defaults to jnn.relu.
            invert: (bool): Whether to invert the bijection. Broadly, True will
                prioritise a faster `inverse` methods, leading to faster `log_prob`,
                False will prioritise faster `transform` methods, leading to faster
                `sample`. Defaults to True
        """
        if base_dist.ndim != 1:
            raise ValueError(f"Expected base_dist.ndim==1, got {base_dist.ndim}")

        dim = base_dist.shape[0]
        permute_strategy = _default_permute_strategy(dim)

        def make_layer(key):  # coupling layer + permutation
            c_key, p_key = random.split(key)
            coupling = Coupling(
                key=c_key,
                transformer=transformer,
                untransformed_dim=dim // 2,
                dim=dim,
                cond_dim=cond_dim,
                nn_width=nn_width,
                nn_depth=nn_depth,
                nn_activation=nn_activation,
            )

            if permute_strategy == "flip":
                return Chain([coupling, Flip((dim,))])
            if permute_strategy == "random":
                perm = Permute(random.permutation(p_key, jnp.arange(dim)))
                return Chain([coupling, perm])
            return coupling

        keys = random.split(key, flow_layers)
        layers = eqx.filter_vmap(make_layer)(keys)
        bijection = Invert(Scan(layers)) if invert else Scan(layers)

        self.nn_width = nn_width
        self.nn_depth = nn_depth
        self.flow_layers = flow_layers
        self.permute_strategy = permute_strategy
        self.shape = (dim,)
        self.cond_shape = None if cond_dim is None else (cond_dim,)
        super().__init__(base_dist, bijection)


class MaskedAutoregressiveFlow(Transformed):
    """Masked autoregressive flow (https://arxiv.org/abs/1606.04934,
    https://arxiv.org/abs/1705.07057v4). Parameterises a transformer with an
    autoregressive neural network.
    """

    flow_layers: int
    nn_width: int
    nn_depth: int
    permute_strategy: str

    def __init__(
        self,
        key: KeyArray,
        base_dist: Distribution,
        transformer: Bijection,
        cond_dim: int | None = None,
        flow_layers: int = 8,
        nn_width: int = 40,
        nn_depth: int = 2,
        nn_activation: Callable = jnn.relu,
        invert: bool = True,
    ):
        """
        Args:
            key (KeyArray): Random seed.
            base_dist (Distribution): Base distribution.
            transformer (Bijection): Bijection parameterised by autoregressive network.
            cond_dim (int): _description_. Defaults to 0.
            flow_layers (int): Number of flow layers. Defaults to 5.
            nn_width (int): Number of hidden layers in neural network. Defaults to 40.
            nn_depth (int): Depth of neural network. Defaults to 2.
            nn_activation (Callable): _description_. Defaults to jnn.relu.
            invert (bool): Whether to invert the bijection. Broadly, True will
                prioritise a faster inverse, leading to faster `log_prob`, False will
                prioritise faster forward, leading to faster `sample`. Defaults to True.
        """
        if len(base_dist.shape) != 1:
            raise ValueError(f"Expected base_dist.ndim==1, got {base_dist.ndim}")

        dim = base_dist.shape[0]
        permute_strategy = _default_permute_strategy(dim)

        def make_layer(key):  # masked autoregressive layer + permutation
            masked_auto_key, p_key = random.split(key)
            masked_autoregressive = MaskedAutoregressive(
                key=masked_auto_key,
                transformer=transformer,
                dim=dim,
                cond_dim=cond_dim,
                nn_width=nn_width,
                nn_depth=nn_depth,
                nn_activation=nn_activation,
            )
            if permute_strategy == "flip":
                return Chain([masked_autoregressive, Flip((dim,))])
            if permute_strategy == "random":
                perm = Permute(random.permutation(p_key, jnp.arange(dim)))
                return Chain([masked_autoregressive, perm])
            return masked_autoregressive

        keys = random.split(key, flow_layers)
        layers = eqx.filter_vmap(make_layer)(keys)
        bijection = Invert(Scan(layers)) if invert else Scan(layers)

        self.nn_width = nn_width
        self.nn_depth = nn_depth
        self.flow_layers = flow_layers
        self.permute_strategy = permute_strategy
        self.shape = (dim,)
        self.cond_shape = None if cond_dim is None else (cond_dim,)
        super().__init__(base_dist, bijection)


class BlockNeuralAutoregressiveFlow(Transformed):
    """Block neural autoregressive flow (BNAF) (https://arxiv.org/abs/1904.04676)."""

    flow_layers: int
    nn_block_dim: int
    nn_depth: int
    permute_strategy: str

    def __init__(
        self,
        key: KeyArray,
        base_dist: Distribution,
        cond_dim: int | None = None,
        nn_depth: int = 1,
        nn_block_dim: int = 8,
        flow_layers: int = 1,
        invert: bool = True,
    ):
        """
        Args:
            key (KeyArray): Jax PRNGKey.
            base_dist (Distribution): Base distribution.
            cond_dim (int | None): Dimension of conditional variables.
            nn_depth (int): Number of hidden layers within the networks.
                Defaults to 1.
            nn_block_dim (int): Block size. Hidden layer width is
                dim*nn_block_dim. Defaults to 8.
            flow_layers (int): Number of BNAF layers. Defaults to 1.
            invert: (bool): Use `True` for access of `log_prob` only (e.g.
                fitting by maximum likelihood), `False` for the forward direction
                (sampling) only (e.g. for fitting variationally).
        """
        if len(base_dist.shape) != 1:
            raise ValueError(f"Expected base_dist.ndim==1, got {base_dist.ndim}")

        dim = base_dist.shape[-1]
        permute_strategy = _default_permute_strategy(base_dist.shape[0])

        def make_layer(key):  # masked autoregressive layer + permutation
            ban_key, p_key = random.split(key)
            ban = BlockAutoregressiveNetwork(
                key=ban_key,
                dim=dim,
                cond_dim=cond_dim,
                depth=nn_depth,
                block_dim=nn_block_dim,
            )
            if permute_strategy == "flip":
                return Chain([ban, Flip((dim,))])
            if permute_strategy == "random":
                perm = Permute(random.permutation(p_key, jnp.arange(dim)))
                return Chain([ban, perm])
            return ban

        keys = random.split(key, flow_layers)
        layers = eqx.filter_vmap(make_layer)(keys)
        bijection = Invert(Scan(layers)) if invert else Scan(layers)

        self.nn_block_dim = nn_block_dim
        self.nn_depth = nn_depth
        self.flow_layers = flow_layers
        self.permute_strategy = permute_strategy

        self.shape = (dim,)
        self.cond_shape = None if cond_dim is None else (cond_dim,)
        super().__init__(base_dist, bijection)


class TriangularSplineFlow(Transformed):
    """A stack of layers, where each layer consists of a triangular affine
    transformation with weight normalisation, and an elementwise rational quadratic
    spline. Tanh is used to constrain to the input to [-1, 1] before spline
    transformations.
    """

    flow_layers: int
    permute_strategy: str
    knots: int
    tanh_max_val: float

    def __init__(
        self,
        key: KeyArray,
        base_dist: Distribution,
        cond_dim: int | None = None,
        flow_layers: int = 8,
        knots: int = 8,
        tanh_max_val: float = 3.0,
        invert: bool = True,
        init: Callable = glorot_uniform(),
    ):
        """
        Args:
            key (KeyArray): Jax random seed.
            base_dist (Distribution): Base distribution of the flow.
            cond_dim (int | None): The number of conditioning features.
                Defaults to None.
            flow_layers (int): Number of flow layers. Defaults to 8.
            knots (int): Number of knots in the splines. Defaults to 8.
            tanh_max_val (float): Maximum absolute value beyond which we use linear
                "tails" in the tanh function. Defaults to 3.0.
            invert: (bool): Use `True` for access of `log_prob` only (e.g.
                fitting by maximum likelihood), `False` for the forward direction
                (sampling) only (e.g. for fitting variationally).
            init (Callable): Initialisation method for the lower triangular weights.
                Defaults to glorot_uniform().
        """
        if len(base_dist.shape) != 1:
            raise ValueError(f"Expected base_dist.ndim==1, got {base_dist.ndim}")

        dim = base_dist.shape[-1]
        permute_strategy = _default_permute_strategy(dim)

        def make_layer(key):
            lt_key, perm_key, cond_key = random.split(key, 3)
            weights = init(lt_key, (dim, dim))
            lt_weights = weights.at[jnp.diag_indices(dim)].set(1)
            lower_tri = TriangularAffine(
                jnp.zeros(dim), lt_weights, weight_normalisation=True
            )

            bijections = [
                TanhLinearTails(tanh_max_val, (dim,)),
                RationalQuadraticSpline(knots, interval=1, shape=(dim,)),
                Invert(TanhLinearTails(tanh_max_val, (dim,))),
                lower_tri,
            ]

            if cond_dim is not None:
                linear_condition = AdditiveCondition(
                    Linear(cond_dim, dim, use_bias=False, key=cond_key),
                    (dim,),
                    (cond_dim,),
                )
                bijections.append(linear_condition)

            if permute_strategy == "flip":
                bijections.append(Flip((dim,)))
            elif permute_strategy == "random":
                bijections.append(
                    Permute(random.permutation(perm_key, jnp.arange(dim)))
                )
            return Chain(bijections)

        keys = random.split(key, flow_layers)
        layers = eqx.filter_vmap(make_layer)(keys)
        bijection = Invert(Scan(layers)) if invert else Scan(layers)

        self.flow_layers = flow_layers
        self.permute_strategy = permute_strategy
        self.knots = knots
        self.tanh_max_val = tanh_max_val
        self.shape = (dim,)
        self.cond_shape = None if cond_dim is None else (cond_dim,)

        super().__init__(base_dist, bijection)


def _default_permute_strategy(dim):
    if dim <= 2:
        return {1: "none", 2: "flip"}[dim]
    return "random"



"""Utility functions for training."""
from functools import partial
from typing import Any, Sequence

import equinox as eqx
import jax.numpy as jnp
import jax.random as jr
import optax
from jax import Array, jit

from flowjax.distributions import Distribution
from flowjax.train.losses import Loss

PyTree = Any


@eqx.filter_jit
def step(
    optimizer: optax.GradientTransformation,
    opt_state: PyTree,
    loss_fn: Loss,
    params: Distribution,
    *args,
    **kwargs
):
    """Carry out a training step with *args and **kwargs passed to the loss function."""
    loss_val, grads = eqx.filter_value_and_grad(loss_fn)(params, *args, **kwargs)
    updates, opt_state = optimizer.update(grads, opt_state)
    params = eqx.apply_updates(params, updates)
    return params, opt_state, loss_val


def train_val_split(key: jr.KeyArray, arrays: Sequence[Array], val_prop: float = 0.1):
    """Random train validation split for a sequence of arrays.

    Args:
        key (KeyArray): Jax random key.
        arrays Sequence[Array]: Sequence of arrays, with matching size on axis 0.
        val_prop (float): Proportion of data to use for validation. Defaults to 0.1.

    Returns:
        tuple[tuple]: (train_arrays, validation_arrays)
    """
    if not 0 <= val_prop <= 1:
        raise ValueError("val_prop should be between 0 and 1.")

    num_samples = arrays[0].shape[0]
    if not all(arr.shape[0] == num_samples for arr in arrays):
        raise ValueError("Array dimensions must match along axis 0.")

    n_train = num_samples - round(val_prop * num_samples)
    arrays = [jr.permutation(key, a) for a in arrays]
    train_arrays = [arr[:n_train] for arr in arrays]
    val_arrays = [arr[n_train:] for arr in arrays]
    return train_arrays, val_arrays


@partial(jit, static_argnums=1)
def get_batches(arrays: Sequence[Array], batch_size: int):
    """Reshape a sequence of arrays to have an additional dimension for the batches,
    i.e. transforming shape ``(data_len, *rest)`` to ``(num_batches, batch_size,
    *rest)``.

    Considerations:
        - The values in the last batch are dropped if truncated, i.e. if
            ``data_len % batch_size != 0``.
        - If the batch size is greater than the length of the data, then we set the
            batch size to be equal to the data length.

    Args:
        arrays (Sequence[Array]): Sequence of arrays, with shape matching on axis 0.
        batch_size (int): The batch size.
    """
    data_len = arrays[0].shape[0]
    if not all(arr.shape[0] == data_len for arr in arrays):
        raise ValueError("Array dimensions do not match along the batch axis.")

    return tuple(_add_batch(arr, batch_size) for arr in arrays)


def _add_batch(arr, batch_size):
    "Adds a leading dimension for batches, dropping the last batch if truncated."
    batch_size = min(batch_size, arr.shape[0])
    n_batches = arr.shape[0] // batch_size
    arr = arr[: n_batches * batch_size].reshape(n_batches, batch_size, *arr.shape[1:])
    return arr


def count_fruitless(losses: list[float]) -> int:
    """Given a list of losses from each epoch, count the number of epochs since
    the minimum loss.

    Args:
        losses (list[float]): List of losses.

    """
    min_idx = jnp.argmin(jnp.array(losses)).item()
    return len(losses) - min_idx - 1



"""Rational quadratic spline bijections (https://arxiv.org/abs/1906.04032)."""


import equinox as eqx
import jax
import jax.numpy as jnp
from jax import Array

from flowjax.bijections.bijection import Bijection
from flowjax.bijections.jax_transforms import Batch
from flowjax.utils import real_to_increasing_on_interval


class _ScalarRationalQuadraticSpline(Bijection):
    """Scaler RationalQuadraticSpline transformation (https://arxiv.org/abs/1906.04032)."""

    knots: int
    interval: float
    softmax_adjust: float
    min_derivative: float
    unbounded_x_pos: Array
    unbounded_y_pos: Array
    unbounded_derivatives: Array

    def __init__(
        self,
        knots: int,
        interval: float,
        min_derivative: float = 1e-3,
        softmax_adjust: float = 1e-2,
    ):
        self.knots = knots
        self.interval = interval
        self.softmax_adjust = softmax_adjust
        self.min_derivative = min_derivative
        self.shape = ()
        self.cond_shape = None

        # Inexact arrays
        self.unbounded_x_pos = jnp.zeros(knots)
        self.unbounded_y_pos = jnp.zeros(knots)
        self.unbounded_derivatives = jnp.full(
            knots + 2, jnp.log(jnp.exp(1 - min_derivative) - 1)
        )

    @property
    def x_pos(self):
        """Get the knot x positions."""
        x_pos = real_to_increasing_on_interval(
            self.unbounded_x_pos, self.interval, self.softmax_adjust
        )
        return jnp.pad(x_pos, 1, constant_values=(-self.interval, self.interval))

    @property
    def y_pos(self):
        """Get the knot y positions."""
        y_pos = real_to_increasing_on_interval(
            self.unbounded_y_pos, self.interval, self.softmax_adjust
        )
        return jnp.pad(y_pos, 1, constant_values=(-self.interval, self.interval))

    @property
    def derivatives(self):
        """Get the knot derivitives."""
        return jax.nn.softplus(self.unbounded_derivatives) + self.min_derivative

    def transform(self, x, condition=None):
        # Following notation from the paper
        # pylint: disable=C0103
        x_pos, y_pos, derivatives = self.x_pos, self.y_pos, self.derivatives
        in_bounds = jnp.logical_and(x > -self.interval, x < self.interval)
        x_robust = jnp.where(in_bounds, x, 0)  # To avoid nans
        k = jnp.searchsorted(x_pos, x_robust) - 1  # k is bin number
        xi = (x_robust - x_pos[k]) / (x_pos[k + 1] - x_pos[k])
        sk = (y_pos[k + 1] - y_pos[k]) / (x_pos[k + 1] - x_pos[k])
        dk, dk1, yk, yk1 = derivatives[k], derivatives[k + 1], y_pos[k], y_pos[k + 1]
        num = (yk1 - yk) * (sk * xi ** 2 + dk * xi * (1 - xi))
        den = sk + (dk1 + dk - 2 * sk) * xi * (1 - xi)
        y = yk + num / den  # eq. 4
        return jnp.where(in_bounds, y, x)

    def transform_and_log_det(self, x, condition=None):
        y = self.transform(x)
        derivative = self.derivative(x)
        return y, jnp.log(derivative).sum()

    def inverse(self, y, condition=None):
        # Following notation from the paper
        # pylint: disable=C0103
        x_pos, y_pos, derivatives = self.x_pos, self.y_pos, self.derivatives
        in_bounds = jnp.logical_and(y > -self.interval, y < self.interval)

        y_robust = jnp.where(in_bounds, y, 0)  # To avoid nans
        k = jnp.searchsorted(y_pos, y_robust) - 1
        xk, xk1, yk, yk1 = x_pos[k], x_pos[k + 1], y_pos[k], y_pos[k + 1]
        sk = (yk1 - yk) / (xk1 - xk)
        y_delta_s_term = (y_robust - yk) * (
            derivatives[k + 1] + derivatives[k] - 2 * sk
        )
        a = (yk1 - yk) * (sk - derivatives[k]) + y_delta_s_term
        b = (yk1 - yk) * derivatives[k] - y_delta_s_term
        c = -sk * (y_robust - yk)
        sqrt_term = jnp.sqrt(b ** 2 - 4 * a * c)
        xi = (2 * c) / (-b - sqrt_term)
        x = xi * (xk1 - xk) + xk
        return jnp.where(in_bounds, x, y)

    def inverse_and_log_det(self, y, condition=None):
        x = self.inverse(y)
        derivative = self.derivative(x)
        return x, -jnp.log(derivative).sum()

    def derivative(self, x) -> Array:
        """The derivative dy/dx of the forward transformation."""
        # Following notation from the paper (eq. 5)
        # pylint: disable=C0103
        x_pos, y_pos, derivatives = self.x_pos, self.y_pos, self.derivatives
        in_bounds = jnp.logical_and(x > -self.interval, x < self.interval)
        x_robust = jnp.where(in_bounds, x, 0)  # To avoid nans
        k = jnp.searchsorted(x_pos, x_robust) - 1
        xi = (x_robust - x_pos[k]) / (x_pos[k + 1] - x_pos[k])
        sk = (y_pos[k + 1] - y_pos[k]) / (x_pos[k + 1] - x_pos[k])
        dk, dk1 = derivatives[k], derivatives[k + 1]
        num = sk ** 2 * (dk1 * xi ** 2 + 2 * sk * xi * (1 - xi) + dk * (1 - xi) ** 2)
        den = (sk + (dk1 + dk - 2 * sk) * xi * (1 - xi)) ** 2
        derivative = num / den
        return jnp.where(in_bounds, derivative, 1.0)  # type: ignore


class RationalQuadraticSpline(Batch):
    """Elementwise rational quadratic spline transform (https://arxiv.org/abs/1906.04032),
    initialised at the identity function.
    """

    def __init__(
        self,
        knots: int,
        interval: float,
        shape: tuple[int, ...] = (),
        min_derivative: float = 1e-3,
        softmax_adjust: float = 1e-2,
    ) -> None:
        """
        Args:
            knots (int): Number of knots.
            interval (float): interval to transform, [-interval, interval].
            shape (tuple[int, ...]): Shape of transformation. Defaults to ().
            min_derivative (float): Minimum dervivative. Defaults to 1e-3.
            softmax_adjust (float): Controls minimum bin width and height by
                rescaling softmax output, e.g. 0=no adjustment, 1=average softmax output
                with evenly spaced widths, >1 promotes more evenly spaced widths.
                See ``real_to_increasing_on_interval``.. Defaults to 1e-2.
        """

        def constructor():
            return _ScalarRationalQuadraticSpline(
                knots, interval, min_derivative, softmax_adjust
            )

        # Create constructor with appropriate number of batch dimensions
        for dim in reversed(shape):
            constructor = eqx.filter_vmap(constructor, axis_size=dim)

        spline = constructor()
        super().__init__(spline, shape, vectorize_bijection=True)



"""Module contains the Chain bijection, that allows sequentially application of
arbitrary bijections, with compatible shapes.
"""
from typing import Sequence

from flowjax.bijections.bijection import Bijection
from flowjax.utils import check_shapes_match, merge_cond_shapes


class Chain(Bijection):
    """Chain together arbitrary bijections to form another bijection."""

    bijections: tuple[Bijection]

    def __init__(self, bijections: Sequence[Bijection]):
        """
        Args:
            bijections (Sequence[Bijection]): Sequence of bijections.
        """
        check_shapes_match([b.shape for b in bijections])
        self.shape = bijections[0].shape
        self.cond_shape = merge_cond_shapes([b.cond_shape for b in bijections])
        self.bijections = tuple(bijections)

    def transform(self, x, condition=None):
        for bijection in self.bijections:
            x = bijection.transform(x, condition)
        return x

    def transform_and_log_det(self, x, condition=None):
        log_abs_det_jac = 0
        for bijection in self.bijections:
            x, log_abs_det_jac_i = bijection.transform_and_log_det(x, condition)
            log_abs_det_jac += log_abs_det_jac_i.sum()
        return x, log_abs_det_jac

    def inverse(self, y, condition=None):
        for bijection in reversed(self.bijections):
            y = bijection.inverse(y, condition)
        return y

    def inverse_and_log_det(self, y, condition=None):
        log_abs_det_jac = 0
        for bijection in reversed(self.bijections):
            y, log_abs_det_jac_i = bijection.inverse_and_log_det(y, condition)
            log_abs_det_jac += log_abs_det_jac_i.sum()
        return y, log_abs_det_jac

    def __getitem__(self, i: int | slice) -> Bijection:
        if isinstance(i, int):
            return self.bijections[i]
        if isinstance(i, slice):
            return Chain(self.bijections[i])
        raise TypeError(f"Indexing with type {type(i)} is not supported.")

    def __iter__(self):
        yield from self.bijections

    def __len__(self):
        return len(self.bijections)



"""SoftPlus bijection"""
import jax.numpy as jnp
from jax.nn import softplus

from flowjax.bijections.bijection import Bijection


class SoftPlus(Bijection):
    r"""Transforms to positive domain using softplus :math:`y = \log(1 + \exp(x))`."""

    def __init__(self, shape: tuple[int] = ()):
        self.shape = shape
        self.cond_shape = None

    def transform(self, x, condition=None):
        x, _ = self._argcheck_and_cast(x)
        return softplus(x)

    def transform_and_log_det(self, x, condition=None):
        x, _ = self._argcheck_and_cast(x)
        return softplus(x), -softplus(-x).sum()

    def inverse(self, y, condition=None):
        y, _ = self._argcheck_and_cast(y)
        return jnp.log(-jnp.expm1(-y)) + y

    def inverse_and_log_det(self, y, condition=None):
        y, _ = self._argcheck_and_cast(y)
        x = self.inverse(y)
        return x, softplus(-x).sum()



"""Example tasks"""

import jax.numpy as jnp
import jax.random as jr


def two_moons(key, n_samples, noise_std=0.2):
    """Two moon distribution."""
    angle_key, noise_key = jr.split(key)
    angle = jr.uniform(angle_key, (n_samples,)) * 2 * jnp.pi
    x = 2 * jnp.cos(angle)
    y = 2 * jnp.sin(angle)
    x = jnp.where(angle > jnp.pi, x + 1, x - 1)
    noise = jr.normal(noise_key, (n_samples, 2)) * noise_std
    return jnp.stack([x, y], axis=1) + noise



"""Affine bijections."""

from typing import Callable

import jax.numpy as jnp
from jax import Array
from jax.experimental import checkify
from jax.scipy.linalg import solve_triangular
from jax.typing import ArrayLike

from flowjax.bijections.bijection import Bijection
from flowjax.bijections.softplus import SoftPlus
from flowjax.utils import arraylike_to_array


class Affine(Bijection):
    """Elementwise affine transformation ``y = a*x + b``. loc and scale should broadcast
    to the desired shape of the bijection.
    """

    loc: Array
    _scale: Array
    positivity_constraint: Bijection

    def __init__(
        self,
        loc: ArrayLike = 0,
        scale: ArrayLike = 1,
        positivity_constraint: Bijection | None = None,
    ):
        """
        Args:
            loc (ArrayLike): Location parameter. Defaults to 0.
            scale (ArrayLike): Scale parameter. Defaults to 1.
            postivity_constraint (Bijection): Bijection with shape matching the Affine
                bijection, that maps the scale parameter from an unbounded domain to the
                positive domain. Defaults to SoftPlus.
        """
        loc, scale = [arraylike_to_array(a, dtype=float) for a in (loc, scale)]
        self.shape = jnp.broadcast_shapes(loc.shape, scale.shape)
        self.cond_shape = None

        self.loc = jnp.broadcast_to(loc, self.shape)

        if positivity_constraint is None:
            positivity_constraint = SoftPlus(self.shape)

        self.positivity_constraint = positivity_constraint
        self._scale = positivity_constraint.inverse(jnp.broadcast_to(scale, self.shape))

    def transform(self, x, condition=None):
        x, _ = self._argcheck_and_cast(x)
        return x * self.scale + self.loc

    def transform_and_log_det(self, x, condition=None):
        x, _ = self._argcheck_and_cast(x)
        scale = self.scale
        return x * scale + self.loc, jnp.log(scale).sum()

    def inverse(self, y, condition=None):
        y, _ = self._argcheck_and_cast(y)
        return (y - self.loc) / self.scale

    def inverse_and_log_det(self, y, condition=None):
        y, _ = self._argcheck_and_cast(y)
        scale = self.scale
        return (y - self.loc) / scale, -jnp.log(scale).sum()

    @property
    def scale(self):
        """The scale parameter of the affine transformation."""
        return self.positivity_constraint.transform(self._scale)


class TriangularAffine(Bijection):
    r"""Transformation of the form :math:`Ax + b`, where :math:`A` is a lower or upper
    triangular matrix."""

    loc: Array
    diag_idxs: Array
    tri_mask: Array
    lower: bool
    positivity_constraint: Bijection
    _arr: Array
    _diag: Array
    _weight_scale: Array | None

    def __init__(
        self,
        loc: ArrayLike,
        arr: ArrayLike,
        lower: bool = True,
        weight_normalisation: bool = False,
        positivity_constraint: Bijection | None = None,
    ):
        """
        Args:
            loc (ArrayLike): Location parameter.
            arr (ArrayLike): Triangular matrix.
            lower (bool): Whether the mask should select the lower or upper
                triangular matrix (other elements ignored). Defaults to True (lower).
            weight_normalisation (bool): If true, carry out weight normalisation.
            postivity_constraint (Bijection): Bijection with shape matching the
                dimension of the triangular affine bijection, that maps the diagonal
                entries of the array from an unbounded domain to the positive domain.
                Also used for weight normalisation parameters, if used. Defaults to
                SoftPlus.
        """
        loc, arr = [arraylike_to_array(a, dtype=float) for a in (loc, arr)]
        if (arr.ndim != 2) or (arr.shape[0] != arr.shape[1]):
            raise ValueError("arr must be a square, 2-dimensional matrix.")
        checkify.check(
            jnp.all(jnp.diag(arr) > 0),
            "arr diagonal entries must be positive",
        )
        dim = arr.shape[0]
        self.diag_idxs = jnp.diag_indices(dim)
        tri_mask = jnp.tril(jnp.ones((dim, dim), dtype=jnp.int32), k=-1)
        self.tri_mask = tri_mask if lower else tri_mask.T
        self.lower = lower

        self.shape = (dim,)
        self.cond_shape = None

        if positivity_constraint is None:
            positivity_constraint = SoftPlus(self.shape)

        self.positivity_constraint = positivity_constraint
        self._diag = positivity_constraint.inverse(jnp.diag(arr))

        # inexact arrays
        self.loc = jnp.broadcast_to(loc, (dim,))
        self._arr = arr

        if weight_normalisation:
            self._weight_scale = positivity_constraint.inverse(jnp.ones((dim,)))
        else:
            self._weight_scale = None

    @property
    def arr(self):
        """Get triangular array, (applies masking, constrains diagonal and weight
        normalisation)."""
        diag = self.positivity_constraint.transform(self._diag)
        off_diag = self.tri_mask * self._arr
        arr = off_diag.at[self.diag_idxs].set(diag)

        if self._weight_scale is not None:
            norms = jnp.linalg.norm(arr, axis=1, keepdims=True)
            scale = self.positivity_constraint.transform(self._weight_scale)[:, None]
            arr = scale * arr / norms

        return arr

    def transform(self, x, condition=None):
        x, _ = self._argcheck_and_cast(x)
        return self.arr @ x + self.loc

    def transform_and_log_det(self, x, condition=None):
        x, _ = self._argcheck_and_cast(x)
        arr = self.arr
        return arr @ x + self.loc, jnp.log(jnp.diag(arr)).sum()

    def inverse(self, y, condition=None):
        y, _ = self._argcheck_and_cast(y)
        return solve_triangular(self.arr, y - self.loc, lower=self.lower)

    def inverse_and_log_det(self, y, condition=None):
        y, _ = self._argcheck_and_cast(y)
        arr = self.arr
        x = solve_triangular(arr, y - self.loc, lower=self.lower)
        return x, -jnp.log(jnp.diag(arr)).sum()


class AdditiveCondition(Bijection):
    """Given a callable ``f``, carries out ``y = x + f(condition)`` as the forward
    transformation and ``x = y - f(condition)`` as the inverse transformation. Note that
    the callable can be a callable module with trainable parameters if desired.

    If used to transform a distribution, this allows the "location" to be changed as a
    function of the conditioning variables.
    """

    module: Callable[[ArrayLike], ArrayLike]

    def __init__(
        self,
        module: Callable[[ArrayLike], ArrayLike],
        shape: tuple[int, ...],
        cond_shape: tuple[int, ...],
    ):
        """
        Args:
            module (Callable[[ArrayLike], ArrayLike]): A callable (e.g. a function or
                callable module) that maps array with shape cond_shape, to a shape
                that is broadcastable with the shape of the bijection.
            shape (tuple[int, ...]): The shape of the bijection.
            cond_shape (tuple[int, ...]): The condition shape of the bijection.

        Example:
            Conditioning using a linear transformation

            .. doctest::

                >>> from flowjax.bijections import AdditiveCondition
                >>> from equinox.nn import Linear
                >>> import jax.numpy as jnp
                >>> import jax.random as jr
                >>> bijection = AdditiveCondition(
                ...     Linear(2, 3, key=jr.PRNGKey(0)), shape=(3,), cond_shape=(2,)
                ...     )
                >>> bijection.transform(jnp.ones(3), condition=jnp.ones(2))
                Array([1.9670618, 0.8156546, 1.7763454], dtype=float32)
        """
        self.module = module
        self.shape = shape
        self.cond_shape = cond_shape

    def transform(self, x, condition=None):
        x, condition = self._argcheck_and_cast(x, condition)
        return x + self.module(condition)  # type: ignore - validated in argcheck

    def transform_and_log_det(self, x, condition=None):
        x, condition = self._argcheck_and_cast(x, condition)
        return self.transform(x, condition), jnp.array(0)

    def inverse(self, y, condition=None):
        y, condition = self._argcheck_and_cast(y, condition)
        return y - self.module(condition)  # type: ignore

    def inverse_and_log_det(self, y, condition=None):
        y, condition = self._argcheck_and_cast(y, condition)
        return self.inverse(y, condition), jnp.array(0)  # type: ignore



"Common loss functions for training normalizing flows."
from typing import Callable

import equinox as eqx
import jax.numpy as jnp
import jax.random as jr
from jax import Array, vmap
from jax.scipy.special import logsumexp
from jax.typing import ArrayLike

from flowjax.distributions import Distribution


class Loss:
    "Base class for loss functions. All losses must implement the loss method."

    @eqx.filter_jit
    def __call__(self, static, params, *args, **kwargs):
        """Combine the partitioned model (see ``equinox.combine`` and
        ``equinox.partition``) and compute the loss."""
        model = eqx.combine(static, params)
        return self.loss(model, *args, **kwargs)

    def loss(self, *args, **kwargs):
        raise NotImplementedError("The loss method must be implemented")


class MaximumLikelihoodLoss(Loss):
    """Loss for fitting a flow with maximum likelihood (negative log likelihood). Can
    be used to learn either conditional or unconditional distributions.

    Args:
        params (Distribution): Trainable parameters for the distribution.
        static (Distribution): Static parameters for the distribution.
        x (Array): Target samples.
        condition (Array | None, optional): Corresponding conditioning variables for
            learning a conditional distribution. Defaults to None.
    """

    @staticmethod
    def loss(dist: Distribution, x: Array, condition: Array | None = None):
        return -dist.log_prob(x, condition).mean()


class ContrastiveLoss(Loss):
    r"""Loss function for use in a sequential neural posterior estimation algorithm.
    Learns a posterior ``p(x|condition)``. Contrastive samples for each ``x`` are
    generated from other x samples in the batch.

    Note, that in a simulation based inference context, often :math:`x` is used to
    denote simulations, and :math:`\theta` for simulation parameters. However, for
    consistency with the rest of the package, we use ``x`` to represent the target
    variable (the simulator parameters), and ``condition`` for the conditioning variable
    (the simulator output/oberved data).

    References:
        - https://arxiv.org/abs/1905.07488
        - https://arxiv.org/abs/2002.03712

    """

    def __init__(self, prior: Distribution, n_contrastive: int):
        """
        Args:
            prior (Distribution): The prior distribution over x (the target variable).
            n_contrastive (int): The number of contrastive samples/atoms to use when
                computing the loss.
        """
        self.prior = prior
        self.n_contrastive = n_contrastive

    def loss(self, dist: Distribution, x: Array, condition: Array):
        contrastive = self._get_contrastive(x)
        joint_log_odds = dist.log_prob(x, condition) - self.prior.log_prob(x)
        contrastive_log_odds = dist.log_prob(
            contrastive, condition
        ) - self.prior.log_prob(contrastive)
        contrastive_log_odds = jnp.clip(contrastive_log_odds, -5)  # Clip for stability
        return -(joint_log_odds - logsumexp(contrastive_log_odds, axis=0)).mean()

    def _get_contrastive(self, theta):
        if theta.shape[0] <= self.n_contrastive:
            raise ValueError(
                f"Number of contrastive samples {self.n_contrastive} must be less than "
                f"the size of theta {theta.shape}."
            )
        # Rolling window over theta batch to create contrastive samples.
        idx = jnp.arange(len(theta))[:, None] + jnp.arange(self.n_contrastive)[None, :]
        contrastive = jnp.roll(theta[idx], -1, axis=0)  # Ensure mismatch with condition
        contrastive = jnp.swapaxes(contrastive, 0, 1)  # (contrastive, batch_size, dim)
        return contrastive


class ElboLoss(Loss):
    """The negative evidence lower bound (ELBO), approximated using samples."""

    target: Callable[[ArrayLike], Array]
    num_samples: int

    def __init__(self, target: Callable[[ArrayLike], Array], num_samples: int):
        """
        Args:
            num_samples (int): Number of samples to use in the ELBO approximation.
            target (Callable[[ArrayLike], Array]): The target, i.e. log posterior
                density up to an additive constant / the negative of the potential
                function, evaluated for a single point.
        """
        self.target = target
        self.num_samples = num_samples

    def loss(self, dist: Distribution, key: jr.KeyArray):
        """Computes an estimate of the negative ELBO loss."""
        samples, log_probs = dist.sample_and_log_prob(key, (self.num_samples,))
        target_density = vmap(self.target)(samples)
        losses = log_probs - target_density
        return losses.mean()



"""Bijections from ``flowjax.bijections``"""

from .affine import AdditiveCondition, Affine, TriangularAffine
from .bijection import Bijection
from .block_autoregressive_network import BlockAutoregressiveNetwork
from .chain import Chain
from .concatenate import Concatenate, Stack
from .coupling import Coupling
from .exp import Exp
from .jax_transforms import Batch, Scan
from .masked_autoregressive import MaskedAutoregressive
from .rational_quadratic_spline import RationalQuadraticSpline
from .softplus import SoftPlus
from .tanh import Tanh, TanhLinearTails
from .utils import EmbedCondition, Flip, Invert, Partial, Permute

__all__ = [
    "AdditiveCondition",
    "Affine",
    "Batch",
    "Bijection",
    "BlockAutoregressiveNetwork",
    "Chain",
    "Concatenate",
    "Coupling",
    "EmbedCondition",
    "Exp",
    "Flip",
    "Invert",
    "MaskedAutoregressive",
    "Partial",
    "Permute",
    "RationalQuadraticSpline",
    "Scan",
    "SoftPlus",
    "Stack",
    "Tanh",
    "TanhLinearTails",
    "TriangularAffine",
]



"""
Various masks, generally used in flows to enforce e.g. a dependency structure
that leads to invertibility and efficient Jacobian determinant calculations.
"""

import jax.numpy as jnp
from jax import Array
from jax.scipy.linalg import block_diag


def rank_based_mask(in_ranks: Array, out_ranks: Array, eq: bool = False):
    """Forms mask matrix, with 1s where the out_ranks > or >= in_ranks.

    Args:
        in_ranks (Array): Ranks of the inputs.
        out_ranks (Array): Ranks of the outputs.
        eq (bool): If true, compares with >= instead of >. Defaults to False.

    Returns:
        Array: Mask with shape `(len(out_ranks), len(in_ranks))`
    """
    assert (in_ranks.ndim) == 1 and (out_ranks.ndim == 1)
    if eq:
        mask = out_ranks[:, None] >= in_ranks
    else:
        mask = out_ranks[:, None] > in_ranks
    return mask.astype(jnp.int32)


def block_diag_mask(block_shape: tuple, n_blocks: int):
    """Block diagonal mask."""
    return block_diag(*jnp.ones((n_blocks, *block_shape), jnp.int32))


def block_tril_mask(block_shape: tuple, n_blocks: int):
    """Upper triangular block mask, excluding diagonal blocks."""
    mask = jnp.zeros((block_shape[0] * n_blocks, block_shape[1] * n_blocks), jnp.int32)
    for i in range(n_blocks):
        mask = mask.at[
            (i + 1) * block_shape[0] :, i * block_shape[1] : (i + 1) * block_shape[1]
        ].set(1)
    return mask



import jax.numpy as jnp

from flowjax.masks import block_diag_mask, block_tril_mask, rank_based_mask


def test_rank_based_mask():
    in_ranks = jnp.arange(2)
    out_ranks = jnp.array([0, 1, 1, 2])

    expected_mask = jnp.array([[0, 0], [1, 0], [1, 0], [1, 1]], dtype=jnp.int32)

    mask = rank_based_mask(in_ranks, out_ranks)
    assert jnp.all(expected_mask == mask)

    in_ranks = jnp.array([0, 0, 1, 1])
    out_ranks = jnp.array([0, 1])

    expected_mask = jnp.array([[0, 0, 0, 0], [1, 1, 0, 0]], dtype=jnp.int32)
    mask = rank_based_mask(in_ranks, out_ranks)
    assert jnp.all(expected_mask == mask)


def test_block_tril_mask():
    args = [(1, 2), 3]
    expected = jnp.array([[0, 0, 0, 0, 0, 0], [1, 1, 0, 0, 0, 0], [1, 1, 1, 1, 0, 0]])
    result = block_tril_mask(*args)
    assert jnp.all(expected == result)


def test_block_diag_mask():
    args = [(1, 2), 3]
    expected = jnp.array([[1, 1, 0, 0, 0, 0], [0, 0, 1, 1, 0, 0], [0, 0, 0, 0, 1, 1]])
    result = block_diag_mask(*args)
    assert jnp.all(expected == result)



"""Basic training script for fitting a flow using variational inference."""
from typing import Any, Callable

import equinox as eqx
import jax.random as jr
import optax
from tqdm import tqdm

from flowjax.distributions import Distribution
from flowjax.train.losses import Loss
from flowjax.train.train_utils import step

PyTree = Any


def fit_to_variational_target(
    key: jr.KeyArray,
    dist: Distribution,
    loss_fn: Loss,
    steps: int = 100,
    learning_rate: float = 5e-4,
    optimizer: optax.GradientTransformation | None = None,
    filter_spec: Callable | PyTree = eqx.is_inexact_array,
    show_progress: bool = True,
):
    """Train a distribution (e.g. a flow) by variational inference to a target
    (e.g. an unnormalized density).

    Args:
        key (jr.KeyArray): Jax PRNGKey.
        dist (Distribution): Distribution object, trainable parameters are found
            using equinox.is_inexact_array.
        loss_fn (Loss | None): The loss function to optimize (e.g. the ElboLoss).
        steps (int, optional): The number of training steps to run. Defaults to 100.
        learning_rate (float, optional): Learning rate. Defaults to 5e-4.
        optimizer (optax.GradientTransformation | None, optional): Optax optimizer. If
            provided, this overrides the default Adam optimizer, and the learning_rate
            is ignored. Defaults to None.
        filter_spec (Callable | PyTree, optional): Equinox `filter_spec` for
            specifying trainable parameters. Either a callable `leaf -> bool`, or a
            PyTree with prefix structure matching `dist` with True/False values.
            Defaults to eqx.is_inexact_array.
        show_progress (bool, optional): Whether to show progress bar. Defaults to True.

    Returns:
        tuple: (distribution, losses).
    """
    if optimizer is None:
        optimizer = optax.adam(learning_rate)

    @eqx.filter_value_and_grad
    def loss_val_and_grad(trainable, static, key):
        model = eqx.combine(trainable, static)
        return loss_fn(model, key)

    params, static = eqx.partition(dist, filter_spec)
    opt_state = optimizer.init(params)

    losses = []

    best_params = params
    keys = tqdm(jr.split(key, steps), disable=not show_progress)

    for key in keys:
        params, opt_state, loss = step(
            optimizer, opt_state, loss_fn, params, static, key
        )
        losses.append(loss.item())
        keys.set_postfix({"loss": loss.item()})
        if loss.item() == min(losses):
            best_params = params

    return eqx.combine(best_params, static), losses



import io
import os
import re

from setuptools import find_packages, setup


def read(filename):
    filename = os.path.join(os.path.dirname(__file__), filename)
    text_type = type("")
    with io.open(filename, mode="r", encoding="utf-8") as fd:
        return re.sub(text_type(r":[a-z]+:`~?(.*?)`"), text_type(r"``\1``"), fd.read())


setup(
    name="flowjax",
    version="10.0.1",
    url="https://github.com/danielward27/flowjax.git",
    license="MIT",
    author="Daniel Ward",
    author_email="danielward27@outlook.com",
    description="Normalizing flow implementations in jax.",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=("tests",)),
    install_requires=[
        "jax",
        "jaxlib>=0.3",
        "equinox>=0.10",
        "tqdm",
        "optax",
    ],
    python_requires=">=3.10",
    extras_require={"dev": ["pytest"]},
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
    ],
)



"""Abstact base classes for the `Bijection` and `Bijection` types. Note when
implementing bijections, by convention we try to i) implement the "transform" methods as
the faster/more intuitive approach (compared to the inverse methods); and ii) implement
only the forward methods if an inverse is not available. The `Invert` bijection can be
used to invert the orientation if a fast inverse is desired (e.g. maximum likelihood
fitting of flows).
"""

from abc import abstractmethod

from equinox import Module
from jax import Array
from jax.typing import ArrayLike

from flowjax.utils import arraylike_to_array


class Bijection(Module):
    """Bijection base class. Similar to :py:class:`~flowjax.distributions.Distribution`,
    bijections have a ``shape`` and a ``cond_shape`` attribute. To allow easy composing
    of bijections, all bijections support passing of conditioning variables (even if
    ignored).

    The methods of bijections do not generally support passing of additional batch
    dimensions, however, ``jax.vmap`` or ``eqx.filter_vmap`` can be used to vmap
    specific methods if desired, and a bijection can be explicitly vectorised using the
    :py:class:`~flowjax.bijections.jax_transforms.Batch` bijection.

    Bijections are registered as Jax PyTrees (as they are equinox modules), so are
    compatible with normal jax operations.

    **Implementing a bijection**

        (1) Inherit from ``Bijection``.
        (2) Define the attributes ``shape`` and ``cond_shape``. A ``cond_shape`` of
            ``None`` is used to represent unconditional bijections.
        (3) Implement the abstract methods ``transform``, ``transform_and_log_det``,
            ``inverse`` and ``inverse_and_log_det``. These should act on
            inputs compatible with the shapes ``shape`` for ``x``, and ``cond_shape``
            for ``condition``.

    """

    shape: tuple[int, ...]
    cond_shape: tuple[int, ...] | None

    @abstractmethod
    def transform(self, x: ArrayLike, condition: ArrayLike | None = None) -> Array:
        """Apply transformation."""

    @abstractmethod
    def transform_and_log_det(
        self, x: ArrayLike, condition: ArrayLike | None = None
    ) -> tuple[Array, Array]:
        """Apply transformation and compute log absolute value of the Jacobian
        determinant."""

    @abstractmethod
    def inverse(self, y: ArrayLike, condition: ArrayLike | None = None) -> Array:
        """Invert the transformation."""

    @abstractmethod
    def inverse_and_log_det(
        self, y: ArrayLike, condition: ArrayLike | None = None
    ) -> tuple[Array, Array]:
        """Invert the transformation and compute log absolute value of the Jacobian
        determinant."""

    def _argcheck_and_cast(
        self, x: ArrayLike, condition: ArrayLike | None = None
    ) -> tuple[Array, Array | None]:
        """Utility function that checks input shapes against the bijection shapes,
        and casts inputs to arrays if required. Note this permits passing a condition
        in the case when bijection.cond_shape is None."""
        x = arraylike_to_array(x, err_name="x")

        if x.shape != self.shape:
            raise ValueError(f"Expected x.shape {self.shape}; got {x.shape}")

        if condition is not None:
            condition = arraylike_to_array(condition, err_name="condition")

            if self.cond_shape is not None and condition.shape != self.cond_shape:
                raise ValueError(
                    f"Expected condition.shape {self.cond_shape}; got {condition.shape}"
                )

        return x, condition
